/* Distributed Checksum Clearinghouse
 *
 * check checksums in the local whitelist
 *
 * Copyright (c) 2005 by Rhyolite Software, LLC
 *
 * This agreement is not applicable to any entity which sells anti-spam
 * solutions to others or provides an anti-spam solution as part of a
 * security solution sold to other entities, or to a private network
 * which employs the DCC or uses data provided by operation of the DCC
 * but does not provide corresponding data to other users.
 *
 * Permission to use, copy, modify, and distribute this software for any
 * purpose with or without fee is hereby granted, provided that the above
 * copyright notice and this permission notice appear in all copies.
 *
 * Parties not eligible to receive a license under this agreement can
 * obtain a commercial license to use DCC and permission to use
 * U.S. Patent 6,330,590 by contacting Commtouch at http://www.commtouch.com/
 * or by email to nospam@commtouch.com.
 *
 * A commercial license would be for Distributed Checksum and Reputation
 * Clearinghouse software.  That software includes additional features.  This
 * free license for Distributed ChecksumClearinghouse Software does not in any
 * way grant permision to use Distributed Checksum and Reputation Clearinghouse
 * software
 *
 * THE SOFTWARE IS PROVIDED "AS IS" AND RHYOLITE SOFTWARE, LLC DISCLAIMS ALL
 * WARRANTIES WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL RHYOLITE SOFTWARE, LLC
 * BE LIABLE FOR ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES
 * OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS,
 * WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION,
 * ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS
 * SOFTWARE.
 *
 * Rhyolite Software DCC 1.3.20-1.94 $Revision$
 */

#include "dcc_ck.h"


#define WHITE_MAGIC_B_STR "DCC client whitelist hash table version "
#define WHITE_MAGIC_V_STR "16"
static const DCC_WHITE_MAGIC white_magic = WHITE_MAGIC_B_STR WHITE_MAGIC_V_STR;

#define EMPTY_WHITE_SIZE    (sizeof(DCC_WHITE_TBL) - sizeof(DCC_WHITE_ENTRY))
#define MAX_WHITE_ENTRIES   (DCC_WHITE_TBL_BINS*10)
#define ENTRIES2SIZE(_l)    (sizeof(DCC_WHITE_TBL)			\
			     - sizeof(DCC_WHITE_ENTRY)			\
			     + ((_l) * sizeof(DCC_WHITE_ENTRY)))



void
dcc_wf_init(DCC_WF *wf,
	    u_int wf_flags)		/* DCC_WF_*/
{
	memset(wf, 0, sizeof(*wf));
	wf->ht_fd = -1;
	wf->wf_flags = wf_flags;
}



/* this is needed only on systems without coherent mmap()/read()/write() */
static void
clean_white(DCC_WF *wf)
{
	if (wf->info
	    && 0 > MSYNC(wf->info, wf->info_size, MS_SYNC))
		dcc_error_msg("msync(%s): %s", wf->ht_nm, ERROR_STR());
}



static void
unmap_white_ht(DCC_WF *wf)
{
	if (!wf->info)
		return;

	clean_white(wf);
#ifdef DCC_WIN32
	win32_unmap(&wf->ht_map, wf->info, wf->ht_nm);
#else
	if (0 > munmap((void *)wf->info, wf->info_size))
		dcc_error_msg("munmap(%s,%d): %s",
			      wf->ht_nm, wf->info_size, ERROR_STR());
#endif
	wf->info = 0;
}



static u_char				/* 1=done, 0=failed */
map_white_ht(DCC_EMSG emsg, DCC_WF *wf, DCC_WHITE_INX entries)
{
	int size;
#ifndef DCC_WIN32
	void *p;
#endif

	unmap_white_ht(wf);

	if (entries > MAX_WHITE_ENTRIES) {
		dcc_pemsg(EX_IOERR, emsg, "%s should not contain %d entries",
			  wf->ht_nm, entries);
		return 0;
	}

#ifdef DCC_WIN32
	/* Make the hash table files maximum size on WIN32.
	 * You cannot change the size of a WIN32 mapping object without
	 * getting all processes using it to release it so that it can be
	 * recreated.  This may cause problems if the size of hash table
	 * header changes.
	 * Since the file does not change size, there is no need to remap it */
	size = ENTRIES2SIZE(MAX_WHITE_ENTRIES);
	if (!wf->info) {
		wf->info = win32_map(emsg, &wf->ht_map, wf->ht_nm, wf->ht_fd,
				     size);
		if (!wf->info)
			return 0;
	}
#else
	size = ENTRIES2SIZE(entries);
	p = mmap(0, size,
		 (wf->wf_flags & DCC_WF_RO)
		 ? PROT_READ : (PROT_READ|PROT_WRITE),
		 MAP_SHARED, wf->ht_fd, 0);
	if (p == MAP_FAILED) {
		dcc_pemsg(EX_IOERR, emsg, "mmap(%s,%d): %s",
			  wf->ht_nm, size, ERROR_STR());
		return 0;
	}
	wf->info = p;
#endif
	wf->info_size = size;
	wf->info_entries = entries;
	wf->info_flags = wf->info->hdr.flags;

	return 1;
}



static u_char
close_white_ht(DCC_EMSG emsg, DCC_WF *wf)
{
	u_char result = 1;

	if (wf->ht_fd < 0)
		return result;

	unmap_white_ht(wf);

#ifdef DCC_WIN32
	/* unlock the file before closing it to keep Win95 happy */
	if (!dcc_unlock_fd(emsg, wf->ht_fd, DCC_LOCK_ALL_FILE,
			   "whitelist ", wf->ht_nm))
		result = 0;
#endif
	if (0 > close(wf->ht_fd)) {
		dcc_pemsg(EX_IOERR, emsg, "close(%s): %s",
			  wf->ht_nm, ERROR_STR());
		result = 0;
	}

	wf->ht_fd = -1;
	return result;
}



static int				/* 0=ok -1=failed */
unlink_white_ht(DCC_WF *wf,
		time_t now)		/* 0=our own file */
{
	int result = 0;

	/* mark it bad if it is a hash table */
	if (!now && wf->info && !(wf->wf_flags & DCC_WF_RO))
		wf->info->hdr.ascii_mtime = 0;

#ifdef DCC_WIN32
	/* racy but you cannot unlink an open file on WIN32 */
	close_white_ht(0, wf);
#endif
	if ((wf->wf_flags & DCC_WF_RO)
	    || 0 > unlink(wf->ht_nm)) {
		if (now
		    && DCC_IS_TIME(now, wf->broken, DCC_WHITE_PARSE_DELAY)) {
			dcc_error_msg("unlink(%s): %s",
				      wf->ht_nm, ERROR_STR());
			wf->broken = now + DCC_WHITE_PARSE_DELAY;
		}
		result = -1;
	}
#ifndef DCC_WIN32
	close_white_ht(0, wf);
#endif

	return result;
}



static u_char
new_ht_nm(DCC_EMSG emsg, DCC_WF *wf, u_char new)
{
	if (wf->ascii_nm_len >= ISZ(wf->ht_nm) - STRZ(DCC_WHITE_SUFFIX)) {
		dcc_pemsg(EX_NOINPUT, emsg, "bad whitelist file name \"%s\"",
			  wf->ascii_nm);
		wf->broken = time(0) + DCC_WHITE_PARSE_DELAY;
		return 0;
	}

	memcpy(wf->ht_nm, wf->ascii_nm, wf->ascii_nm_len);
	strcpy(&wf->ht_nm[wf->ascii_nm_len],
	       new ? DCC_WHITE_NEW_SUFFIX : DCC_WHITE_SUFFIX);
	return 1;
}



u_char
dcc_new_white_nm(DCC_EMSG emsg, DCC_WF *wf,
		 const char *new_white_nm)
{
	DCC_PATH new_path;
	u_int len;
	int i;

	dcc_fnm2path(new_path, new_white_nm);
	len = strlen(new_path);
	if (wf->ht_fd >= 0
	    && len == wf->ascii_nm_len
	    && !memcmp(new_path, wf->ascii_nm, len))
		return 1;

	close_white_ht(0, wf);
	dcc_wf_init(wf, wf->wf_flags);

	i = len - STRZ(DCC_WHITE_SUFFIX);
	if (i > 0
	    && (!strcmp(&new_path[i], DCC_WHITE_SUFFIX)
		|| !strcmp(&new_path[i], DCC_WHITE_NEW_SUFFIX))) {
		len = i;
		new_path[len] = '\0';
	}

	if (len >= sizeof(wf->ht_nm) - sizeof(DCC_WHITE_SUFFIX)) {
		dcc_pemsg(EX_USAGE, emsg, "long whitelist name \"%s\"",
			  new_path);
		return 0;
	}

	memcpy(wf->ascii_nm, new_path, len+1);
	wf->ascii_nm_len = len;
	return 1;
}



/* open and shared-lock a hash table file */
static int				/* -1=fatal 0=rebuild, 1=ok */
open_white_ht(DCC_EMSG emsg, DCC_WF *wf)
{
	struct stat sb;
	int size;
	DCC_WHITE_INX entries;

	close_white_ht(0, wf);

	if (!new_ht_nm(emsg, wf, 0))
		return -1;

	wf->ht_fd = dcc_lock_open(emsg, wf->ht_nm,
				  (wf->wf_flags & DCC_WF_RO)
				  ? O_RDONLY : O_RDWR,
				  DCC_LOCK_OPEN_SHARE,
				  DCC_LOCK_ALL_FILE, 0);
	if (wf->ht_fd < 0)
		return 0;

	if (fstat(wf->ht_fd, &sb) < 0) {
		dcc_pemsg(EX_IOERR, emsg, "stat(%s): %s",
			  wf->ht_nm, ERROR_STR());
		wf->broken = time(0) + DCC_WHITE_PARSE_DELAY;
		return -1;
	}

	size = sb.st_size - EMPTY_WHITE_SIZE;
	if (size < 0) {
		dcc_pemsg(EX_NOINPUT, emsg,
			  "%s is too small to be a DCC whitelist hash table",
			  wf->ht_nm);
		if (sb.st_size < ISZ(DCC_WHITE_MAGIC)) {
			dcc_pemsg(EX_NOINPUT, emsg,
				  "%s is too small"
				  " to be a DCC whitelist hash table",
				  wf->ht_nm);
			return unlink_white_ht(wf, time(0));
		}
		entries = MAX_WHITE_ENTRIES+1;
		/* temporarily map the file to check the magic string below */
		if (!map_white_ht(emsg, wf, 0))
			return unlink_white_ht(wf, time(0));

	} else {
		entries = size / sizeof(DCC_WHITE_ENTRY);
		if (!map_white_ht(emsg, wf, entries))
			return unlink_white_ht(wf, time(0));
	}

	if (size < 0
	    || memcmp(&wf->info->magic, &white_magic,
		      sizeof(white_magic))) {
		/* rebuild old format files */
		if (!memcmp(&wf->info->magic, WHITE_MAGIC_B_STR,
			    sizeof(WHITE_MAGIC_B_STR)-1))
			dcc_trace_msg("%s is obsolete %s",
				      wf->ht_nm, wf->info->magic);
		else
			dcc_pemsg(EX_NOINPUT, emsg,
				  "%s is not a DCC whitelist hash file",
				  wf->ht_nm);
		return unlink_white_ht(wf, time(0));
	}

	if ((size % sizeof(DCC_WHITE_ENTRY)) != 0
	    || entries > MAX_WHITE_ENTRIES
	    || entries < wf->info->hdr.entries) {
		dcc_pemsg(EX_NOINPUT, emsg,
			  "impossible size of whitelist %s: "OFF_DPAT,
			  wf->ht_nm, sb.st_size);
		return unlink_white_ht(wf, time(0));
	}

	if (wf->info->hdr.ascii_mtime == 0) {
		close_white_ht(0, wf);
		return 0;		/* broken hash table */
	}

	/* we know the hash table is usable
	 * but we might want to rebuild it */
	wf->changed = 0;
	wf->info_flags = wf->info->hdr.flags;

	/* the wlist command works on both per-user and global whitelists */
	if (wf->wf_flags & DCC_WF_WLIST_CMD) {
		if (wf->info_flags & DCC_WHITE_FG_PER_USER)
			wf->wf_flags |= DCC_WF_PER_USER;
		else
			wf->wf_flags &= ~DCC_WF_PER_USER;
	}

	return 1;
}



static int				/* -1 or fd */
create_white_ht_sub(DCC_EMSG emsg, const DCC_WF *wf, DCC_WF *new_wf,
		    u_char *busyp)
{
	struct stat sb;
	int fd;

	/* do not use O_EXCL because we want to wait for any other
	 * process to finish
	 *
	 * wait if and only if we do not have a usable file open */

	fd = dcc_lock_open(emsg, new_wf->ht_nm,
			   O_RDWR | O_CREAT,
			   wf->ht_fd >= 0 ? DCC_LOCK_OPEN_NOWAIT : 0,
			   DCC_LOCK_ALL_FILE, busyp);
	if (fd < 0)
		return fd;

	/* a new hash table must be empty */
	if (0 > fstat(fd, &sb)) {
		dcc_pemsg(EX_IOERR, emsg, "stat(%s): %s",
			  new_wf->ht_nm, ERROR_STR());
		close_white_ht(emsg, new_wf);
		return -1;
	}

	if (sb.st_size != 0) {
		dcc_pemsg(EX_IOERR, emsg, "%s has non-zero size %d",
			  new_wf->ht_nm, (int)sb.st_size);
		close_white_ht(emsg, new_wf);
		return -1;
	}

	return fd;
}



/* create and write-lock a new hash table file
 *	wait for lock if we don't have existing file open */
static int				/* 1=done, 0=file busy, -1=fatal */
create_white_ht(DCC_EMSG emsg, const DCC_WF *wf, DCC_WF *new_wf)
{
	u_char busy = 0;

	close_white_ht(0, new_wf);

	if (new_wf->wf_flags & DCC_WF_RO) {
		dcc_pemsg(EX_IOERR, emsg,
			  "read only access; cannot create whitelist %s",
			  new_wf->ht_nm);
		new_wf->ht_nm[0] = '\0';
		return -1;
	}

	if (!new_ht_nm(emsg, new_wf, 1)) {
		new_wf->ht_nm[0] = '\0';
		return -1;
	}

#ifndef DCC_WIN32
	/* We want to create a private hash table if the ASCII file
	 * is private, but a hash table owned by the DCC user if the
	 * ASCII file is public */
	if (0 > access(new_wf->ascii_nm, R_OK | W_OK)
	    && dcc_get_priv_home(new_wf->ht_nm)) {
		/* first try to open a public hash table */
		new_wf->ht_fd = create_white_ht_sub(emsg, wf, new_wf, &busy);
		if (new_wf->ht_fd < 0 && !busy) {
			if (emsg && dcc_clnt_debug > 2)
				dcc_error_msg("%s", emsg);
			unlink(new_wf->ht_nm);
			new_wf->ht_fd = create_white_ht_sub(emsg, wf, new_wf,
							&busy);
		}
		dcc_rel_priv();
	}
#endif

	if (new_wf->ht_fd < 0 && !busy) {
		/* try to open or create a private hash table */
		new_wf->ht_fd = create_white_ht_sub(emsg, wf, new_wf,
						    &busy);
		if (new_wf->ht_fd < 0 && !busy) {
			if (emsg && dcc_clnt_debug > 2)
				dcc_error_msg("%s", emsg);
			unlink(new_wf->ht_nm);
			new_wf->ht_fd = create_white_ht_sub(emsg, wf, new_wf,
							&busy);
		}
	}

#ifndef DCC_WIN32
	/* try one last time with privileges in case the ASCII file has
	 * mode 666 but the directory does not */
	if (new_wf->ht_fd < 0 && !busy) {
		if (dcc_get_priv_home(new_wf->ht_nm)) {
			if (emsg && dcc_clnt_debug > 2)
				dcc_error_msg("%s", emsg);
			unlink(new_wf->ht_nm);
			new_wf->ht_fd = create_white_ht_sub(emsg, wf, new_wf,
							&busy);
			dcc_rel_priv();
		}
	}
#endif
	if (new_wf->ht_fd < 0) {
		new_wf->ht_nm[0] = '\0';
		if (busy)
			return 0;
		return -1;
	}
	return 1;
}



#define FIND_WHITE_BROKEN ((DCC_WHITE_ENTRY *)-1)
static DCC_WHITE_ENTRY *
find_white(DCC_EMSG emsg, DCC_WF *wf, DCC_CK_TYPES type, const DCC_SUM sum,
	   DCC_WHITE_INX *binp)
{
	u_long accum;
	DCC_WHITE_INX bin, inx;
	DCC_WHITE_ENTRY *e;
	int loop_cnt, i;

	if (wf->info->hdr.ascii_mtime == 0)
		return FIND_WHITE_BROKEN;

	accum = type;
	for (i = sizeof(DCC_SUM)-1; i >= 0; --i)
		accum = (accum >> 28) + (accum << 4) + sum[i];
	bin = accum % DIM(wf->info->bins);
	if (binp)
		*binp = bin;
	inx = wf->info->bins[bin];

	for (loop_cnt = wf->info->hdr.entries+1;
	     loop_cnt >= 0;
	     --loop_cnt) {
		if (!inx)
			return 0;
		--inx;
		/* if necessary, expand the mapped window into the file */
		if (inx >= wf->info_entries) {
			if (inx >= wf->info->hdr.entries) {
				dcc_pemsg(EX_DATAERR, emsg,
					  "bogus index %u in %s",
					  inx, wf->ht_nm);
				if (!(wf->wf_flags & DCC_WF_RO))
					wf->info->hdr.ascii_mtime = 0;
				clean_white(wf);
				return FIND_WHITE_BROKEN;
			}
			if (!map_white_ht(emsg, wf, wf->info->hdr.entries))
				return 0;
		}
		e = &wf->info->tbl[inx];
		if (e->type == type && !memcmp(e->sum, sum, sizeof(DCC_SUM)))
			return e;
		inx = e->fwd;
	}

	dcc_pemsg(EX_DATAERR, emsg, "chain length %d in %s"
		  " starting at %d near %d for %s %s",
		  wf->info->hdr.entries+1,
		  wf->ht_nm,
		  (DCC_WHITE_INX)(accum % DIM(wf->info->bins)), inx,
		  dcc_type2str_err(type, 0, 0, 0), dcc_ck2str_err(type, sum));

	if (!(wf->wf_flags & DCC_WF_RO))
		wf->info->hdr.ascii_mtime = 0;
	clean_white(wf);
	return FIND_WHITE_BROKEN;
}



static int				/* 1=quit stat_white_nms */
stat_1white_nm(int *resultp, DCC_WF *wf, const char *nm, time_t ascii_mtime)
{
	struct stat sb;

	if (stat(nm, &sb) < 0) {
		time_t now = time(0);
		if (wf->missing_time != 0) {
			*resultp = !DCC_IS_TIME(now, wf->missing_time,
						DCC_WHITE_BRIDGE_SECS);
			return 1;
		}
		dcc_trace_msg("%s disappeared: %s", nm, ERROR_STR());
		wf->missing_time = now + DCC_WHITE_BRIDGE_SECS;
		*resultp = 1;
		return 1;
	}

	if (sb.st_mtime != ascii_mtime) {
		*resultp = 0;
		return 1;
	}

	return 0;
}



/* see if the ASCII files have changed */
static int				/* 1=unchanged 0=changed -1=broken */
stat_white_nms(DCC_EMSG emsg, DCC_WF *wf)
{
	struct stat ht_fd_sb;
#ifndef DCC_WIN32
	struct stat ht_sb;
#endif
	time_t now;
	int i, result;

	if (!wf->info)
		return -1;

	now = time(0);
	wf->last_stat_time = now;
	wf->next_stat_time = now + DCC_WHITE_STAT_DELAY;

	/* Notice if the hash file has been unlinked */
	if (fstat(wf->ht_fd, &ht_fd_sb) < 0) {
		dcc_pemsg(EX_IOERR, emsg, "stat(%s): %s",
			  wf->ht_nm, ERROR_STR());
		return -1;
	}
#ifndef DCC_WIN32	/* open files cannot be unlinked in WIN32 */
	if (stat(wf->ht_nm, &ht_sb) < 0) {
		if (emsg && dcc_clnt_debug)
			dcc_error_msg("stat(%s): %s",
				      wf->ht_nm, ERROR_STR());
		return -1;
	}
	if (ht_fd_sb.st_dev != ht_sb.st_dev
	    || ht_fd_sb.st_ino != ht_sb.st_ino) {
		if (emsg && dcc_clnt_debug > 2)
			dcc_error_msg("%s disappeared", wf->ht_nm);
		return -1;
	}
#endif

	/* delays on re-parsing and complaining from the file rule */
	if (wf->reparse < wf->info->hdr.reparse)
		wf->reparse = wf->info->hdr.reparse;
	if (wf->broken < wf->info->hdr.broken)
		wf->broken = wf->info->hdr.broken;

	/* pretend things are fine if they are recently badly broken */
	if (!DCC_IS_TIME(now, wf->broken, DCC_WHITE_PARSE_DELAY))
		return 1;
	/* seriously broken hash tables are unusable */
	if (wf->info->hdr.ascii_mtime == 0)
		return -1;

	/* force reparsing of syntax errors */
	if (wf->reparse != 0
	    && DCC_IS_TIME(now, wf->reparse, DCC_WHITE_PARSE_DELAY))
		return 0;

	/* if the main ASCII file has disappeared,
	 * leave the hash file open and just complain
	 * but only for a while */
	if (stat_1white_nm(&result, wf, wf->ascii_nm,
			   wf->info->hdr.ascii_mtime))
		return result;
	/* see if any of the included ASCII files are new */
	for (i = 0; i < DIM(wf->info->hdr.white_incs); ++i) {
		if (wf->info->hdr.white_incs[i].nm[0] == '\0')
			break;
		/* stop at the first missing or changed included file */
		if (stat_1white_nm(&result, wf,
				   wf->info->hdr.white_incs[i].nm,
				   wf->info->hdr.white_incs[i].mtime))
			return result;
	}
	wf->missing_time = 0;

	if ((wf->info_flags & DCC_WHITE_FG_PER_USER)
	    && !(wf->wf_flags & DCC_WF_PER_USER)) {
		dcc_error_msg("%s is a per-user whitelist"
			      " used as a global whitelist",
			      wf->ht_nm);
		return 0;
	}
	if (!(wf->info_flags & DCC_WHITE_FG_PER_USER)
	    && (wf->wf_flags & DCC_WF_PER_USER)) {
		dcc_error_msg("%s is a global whitelist"
			      " used as a per-user whitelist",
			      wf->ht_nm);
		return 0;
	}

	/* Checksums of SMTP client IP addresses are compared against the
	 * checksums of the IP addresses of the hostnames in the ASCII file.
	 * Occassionaly check for changes in DNS A RR's for entries in
	 * the ASCII file, but only if there are host names or IP
	 * addresses in the file */
	if ((wf->info->hdr.flags & DCC_WHITE_FG_HOSTNAMES)
	    && DCC_IS_TIME(now, ht_fd_sb.st_mtime+DCC_RE_RESOLVE,
			   DCC_RE_RESOLVE)) {
		if (dcc_clnt_debug > 2)
			dcc_trace_msg("time to rebuild %s", wf->ht_nm);
		return 0;
	}

	return 1;
}



static u_char
write_white(DCC_EMSG emsg, DCC_WF *wf, const void *buf, int buf_len, off_t pos)
{
	int i;

	if (wf->info) {
#ifdef DCC_WIN32
		/* Windows disclaims coherence between ordinary writes
		 * and memory mapped writing.  The hash tables are
		 * fixed size on Windows because of problems with WIN32
		 * mapping objects, so we do not need to worry about
		 * extending the hash table file. */
		memcpy((char *)wf->info+pos, buf, buf_len);
		return 1;
#else
		/* some UNIX systems have coherence trouble without msync() */
		if (0 > MSYNC(wf->info, wf->info_size, MS_SYNC)) {
			dcc_pemsg(EX_IOERR, emsg, "msync(%s): %s",
				  wf->ht_nm, ERROR_STR());
			return 0;
		}
#endif
	}

	i = lseek(wf->ht_fd, pos, SEEK_SET);
	if (i < 0) {
		dcc_pemsg(EX_IOERR, emsg, "lseek(%s,"OFF_DPAT"): %s",
			  wf->ht_nm, pos, ERROR_STR());
		return 0;
	}
	i = write(wf->ht_fd, buf, buf_len);
	if (i != buf_len) {
		if (i < 0)
			dcc_pemsg(EX_IOERR, emsg, "write(%s,%d): %s",
				  wf->ht_nm, buf_len, ERROR_STR());
		else
			dcc_pemsg(EX_IOERR, emsg, "write(%s,%d): %d",
				  wf->ht_nm, buf_len, i);
		return 0;
	}
	return 1;
}



static DCC_TGTS
add_white_tgts(DCC_TGTS new, DCC_TGTS old)
{
	if (new < DCC_TGTS_TOO_MANY
	    || old < DCC_TGTS_TOO_MANY)
		return DCC_TGTS_INVALID;

	if (new == DCC_TGTS_TOO_MANY) {
		if (old == DCC_TGTS_TOO_MANY)
			return DCC_TGTS_TOO_MANY;
		return DCC_TGTS_INVALID;
	}

	if (new == DCC_TGTS_OK2) {
		if (old == DCC_TGTS_OK2
		    || old == DCC_TGTS_OK)
			return old;
		if (old == DCC_TGTS_OK_MX
		    || old == DCC_TGTS_OK_MXDCC)
			return new;
		return DCC_TGTS_INVALID;
	}

	if (new == DCC_TGTS_OK) {
		if (old == DCC_TGTS_OK
		    || old == DCC_TGTS_OK2
		    || old == DCC_TGTS_OK_MX
		    || old == DCC_TGTS_OK_MXDCC)
			return new;
		return DCC_TGTS_INVALID;
	}

	if (new == DCC_TGTS_OK_MX) {
		if (old == DCC_TGTS_OK
		    || old == DCC_TGTS_OK2
		    || old == DCC_TGTS_OK_MX
		    || old == DCC_TGTS_OK_MXDCC)
			return old;
		return DCC_TGTS_INVALID;
	}

	if (new == DCC_TGTS_OK_MXDCC) {
		if (old == DCC_TGTS_OK
		    || old == DCC_TGTS_OK2
		    || old == DCC_TGTS_OK_MXDCC)
			return old;
		if (old == DCC_TGTS_OK_MX)
			return new;
		return DCC_TGTS_INVALID;
	}

	return DCC_TGTS_INVALID;
}



static int				/* 1=ok,  0=bad entry, -1=fatal */
add_white(DCC_EMSG emsg, DCC_WF *wf, const char *fnm, int lineno,
	  DCC_CK_TYPES type, DCC_SUM sum, DCC_TGTS tgts)
{
	DCC_WHITE_ENTRY *e, new;
	DCC_WHITE_INX bin;
	off_t end;

	/* find the hash chain for the new entry */
	e = find_white(emsg, wf, type, sum, &bin);
	if (e == FIND_WHITE_BROKEN)
		return -1;
	if (e) {
		/* try to combine duplicates */
		tgts = add_white_tgts(tgts, e->tgts);
		if (tgts != DCC_TGTS_INVALID) {
			e->tgts = tgts;
			return 1;
		}
		dcc_pemsg(EX_DATAERR, emsg,
			  "conflicting count for duplicate entry%s",
			  fnm_lineno(fnm, lineno));
		return 0;
	}

	memset(&new, 0, sizeof(new));
	new.type = type;
	memcpy(new.sum, sum, sizeof(DCC_SUM));
	new.tgts = tgts;
	new.fwd = wf->info->bins[bin];

	/* Use a new entry at the end of the file
	 * It will be beyond the memory mapped window into the file */
	if (wf->info->hdr.entries >= MAX_WHITE_ENTRIES) {
		dcc_pemsg(EX_DATAERR, emsg, "already maximum %d entries%s",
			  wf->info->hdr.entries,
			  fnm_lineno(fnm, lineno));
		return -1;
	}
	end = ENTRIES2SIZE(wf->info->hdr.entries);
	wf->info->bins[bin] = ++wf->info->hdr.entries;
	return write_white(emsg, wf, &new, sizeof(new), end) ? 1 : -1;
}



static int				/* 1=ok,  0=bad entry, -1=fatal */
add_white_cidr(DCC_EMSG emsg, DCC_WHITE_TBL *tbl, const char *fnm, int lineno,
	       int bits,
	       const struct in6_addr *addrp, const struct in6_addr *maskp,
	       DCC_TGTS tgts)
{
	int i;
	DCC_WHITE_CIDR_ENTRY *e;

	i = tbl->hdr.cidr.len;
	if (i >= DIM(tbl->hdr.cidr.e)) {
		dcc_pemsg(EX_DATAERR, emsg, "too many CIDR blocks%s",
			  fnm_lineno(fnm, lineno));
		return 0;
	}

	for (e = tbl->hdr.cidr.e; i > 0; ++e, --i) {
		if (e->bits == bits
		    && !memcmp(addrp, &e->addr, sizeof(*addrp))) {
			/* try to combine duplicates */
			tgts = add_white_tgts(tgts, e->tgts);
			if (tgts != DCC_TGTS_INVALID) {
				e->tgts = tgts;
				return 1;
			}
			dcc_pemsg(EX_DATAERR, emsg,
				  "conflicting duplicate CIDR block%s",
				  fnm_lineno(fnm, lineno));
			return 0;
		}
	}

	e->bits = bits;
	e->addr = *addrp;
	e->mask = *maskp;
	e->tgts = tgts;
	++tbl->hdr.cidr.len;

	return 1;
}



static void
fail_new_wf(DCC_WF *new_wf)
{
	if (new_wf->ht_nm[0] != '\0') {
		unlink_white_ht(new_wf, 0);
		new_wf->ht_nm[0] = '\0';
	}

	new_wf->broken = time(0) + DCC_WHITE_PARSE_DELAY;
}



static int				/* -1=fatal 0=usable file w/problems */
make_white_hash_bail(DCC_WF *wf, DCC_WF *new_wf)
{
	fail_new_wf(new_wf);

	if (wf->info) {
		/* we have a usable file open
		 *
		 * this is racy, but the worst that can happen is that
		 * another process races to set the retry timer */
		if (!(wf->wf_flags & DCC_WF_RO)) {
			wf->info->hdr.broken = wf->broken;
			clean_white(wf);
		}
		return 0;
	}

	return -1;
}



/* (re)create the hash table file */
static int				/* 1=ok 0=complain -1=broken */
make_white_hash(DCC_EMSG emsg, DCC_WF *wf)
{
	static u_char zero = 0;
	DCC_WF new_wf;
	int ascii_fd;
	FILE *ascii_file;
	struct stat ascii_sb;
	int parse_result, i;

	if (dcc_clnt_debug > 1)
		dcc_trace_msg("start parsing %s", wf->ascii_nm);

	/* Do not wait to create a new, locked hash table file
	 *	if we have a usable file.  Assume some other process
	 *	is re-parsing the ASCII file.
	 * If we should wait to create a new file, then we don't have a
	 *	usable hash table and so there is no reason to unlock the
	 *	DCC_WF structure.
	 * There is no need to unlock the DCC_WF structure for per-user
	 *	whiteclnt files, because they contain no host names and
	 *	are quick to parse after we get a new hash table file.
	 * The "resolver thread" that parses and re-resolves host names in
	 *	the global whiteclnt file pretends it does not have an
	 *	alternative.  That fits its lack of need to unlock its fake
	 *	DCC_WF structure. */

	dcc_wf_init(&new_wf, wf->wf_flags);
	new_wf.ascii_nm_len = wf->ascii_nm_len;
	memcpy(new_wf.ascii_nm, wf->ascii_nm, wf->ascii_nm_len+1);

	if (0 > stat(new_wf.ascii_nm, &ascii_sb)) {
		dcc_pemsg(EX_IOERR, emsg, "stat(%s): %s",
			  wf->ascii_nm, ERROR_STR());
		/* Delete the hash table if the ASCII file has disappeared
		 * stat_white_nms() delays forcing a re-build if the ASCII
		 * file disappears only temporarily */
		if (wf->ht_nm[0] != '\0') {
			close_white_ht(0, wf);
			unlink(wf->ht_nm);
			dcc_trace_msg("delete %s after %s disappeared",
				      wf->ht_nm, wf->ascii_nm);
		}
		return make_white_hash_bail(wf, &new_wf);
	}

	i = create_white_ht(emsg, wf, &new_wf);
	if (i == 0)			/* the new file is busy */
		return 0;		/*	so ignore the need to reparse */
	if (i < 0)
		return make_white_hash_bail(wf, &new_wf);

	if (!write_white(emsg, &new_wf, white_magic, sizeof(white_magic), 0)
	    || !write_white(emsg, &new_wf, &zero, 1, EMPTY_WHITE_SIZE-1)
	    || !map_white_ht(emsg, &new_wf, 0))
		return make_white_hash_bail(wf, &new_wf);

	if (new_wf.wf_flags & DCC_WF_PER_USER)
		new_wf.info->hdr.flags |= DCC_WHITE_FG_PER_USER;

	ascii_fd = dcc_lock_open(emsg, new_wf.ascii_nm, O_RDONLY,
				 DCC_LOCK_OPEN_NOWAIT | DCC_LOCK_OPEN_SHARE,
				 DCC_LOCK_ALL_FILE, 0);
	if (ascii_fd == -1)
		return make_white_hash_bail(wf, &new_wf);
	ascii_file = fdopen(ascii_fd, "r");
	if (!ascii_file) {
		dcc_pemsg(EX_IOERR, emsg, "fdopen(%s): %s",
			  wf->ascii_nm, ERROR_STR());
		/* don't bother the hash table for our malloc() problem */
		close(ascii_fd);
		fail_new_wf(&new_wf);
		return -1;
	}

	new_wf.info->hdr.ascii_mtime = ascii_sb.st_mtime;
	parse_result = dcc_parse_whitefile(emsg, &new_wf,
					   new_wf.ascii_nm, ascii_file,
					   add_white, add_white_cidr);
#ifdef DCC_WIN32
	/* unlock the file before closing it to keep Win95 happy */
	dcc_unlock_fd(0, ascii_fd, DCC_LOCK_ALL_FILE,
		      "whitelist ", new_wf.ascii_nm);
#endif
	fclose(ascii_file);
	if (parse_result < 0)
		return make_white_hash_bail(wf, &new_wf);
	if (parse_result == 0) {
		/* ensure continued complaints about errors */
		new_wf.info->hdr.reparse = time(0) + DCC_WHITE_PARSE_DELAY;
	}
	clean_white(&new_wf);

#ifdef DCC_WIN32
	/* WIN32 prohibits renaming open files
	 * and there is little or no concurrency on WIN32 DCC clients
	 * So lock the original file and copy to it. */
	close_white_ht(0, wf);
	wf->ht_fd = dcc_lock_open(emsg, wf->ht_nm, O_RDWR | O_CREAT,
				  0, DCC_LOCK_ALL_FILE, 0);
	if (wf->ht_fd < 0)
		return make_white_hash_bail(wf, &new_wf);
	if (!map_white_ht
	    (emsg, wf, new_wf.info_entries))
		return make_white_hash_bail(wf, &new_wf);
	memcpy(wf->info, new_wf.info, new_wf.info_size);
	close_white_ht(emsg, &new_wf);
	unlink(new_wf.ht_nm);
#else
	i = rename(new_wf.ht_nm, wf->ht_nm);
	if (0 > i
	    && dcc_get_priv_home(wf->ht_nm)) {
		i = rename(new_wf.ht_nm, wf->ht_nm);
		dcc_rel_priv();
	}
	if (0 > i) {
		dcc_pemsg(EX_IOERR, emsg, "rename(%s, %s): %s",
			  new_wf.ht_nm, wf->ht_nm, ERROR_STR());
		return make_white_hash_bail(wf, &new_wf);
	}
#endif
	close_white_ht(0, &new_wf);

	if (dcc_clnt_debug > 1)
		dcc_trace_msg("finished parsing %s", wf->ascii_nm);

	i = open_white_ht(emsg, wf);

	wf->last_stat_time = time(0);
	wf->next_stat_time = wf->last_stat_time + DCC_WHITE_STAT_DELAY;

	return min(parse_result,i);
}



/* see that a local whitelist is ready
 *	on failure the file is not locked but the mutex is always locked */
DCC_WHITE_RESULT
dcc_rdy_white(DCC_EMSG emsg, DCC_WF *wf)
{
	time_t now;
	int i;

	now = time(0);

	if (wf->changed) {
		wf->broken = 0;
		wf->reparse = 0;
		close_white_ht(0, wf);
	}

	if (wf->ht_fd >= 0) {
		if (!DCC_IS_TIME(now, wf->next_stat_time,
					DCC_WHITE_STAT_DELAY))
			return DCC_WHITE_OK;
		i = stat_white_nms(emsg, wf);
		if (i > 0)
			return DCC_WHITE_OK;
		if (i < 0) {
			i = open_white_ht(emsg, wf);
			if (i < 0)
				return DCC_WHITE_COMPLAIN;
			if (i > 0)
				i = stat_white_nms(emsg, wf);
		}

	} else {
		if (wf->ascii_nm[0] == '\0') {
			dcc_pemsg(EX_NOINPUT, emsg, "no whitelist");
			/* no file means nothing is listed */
			return DCC_WHITE_NOFILE;
		}

		/* If things are broken,
		 * and it has not been a while,
		 * then assume things are still broken. */
		if (!DCC_IS_TIME(now, wf->broken, DCC_WHITE_PARSE_DELAY)) {
			dcc_pemsg(EX_DATAERR, emsg,
				  "%s still broken", wf->ascii_nm);
			return (dcc_clnt_debug > 2
				? DCC_WHITE_COMPLAIN
				: DCC_WHITE_SILENT);
		}

		i = open_white_ht(emsg, wf);
		if (i < 0)
			return DCC_WHITE_COMPLAIN;
		if (i > 0)
			i = stat_white_nms(emsg, wf);
	}

	if (i > 0)
		return DCC_WHITE_OK;

	/* try to let the resolver thread wait for the DNS chitchat */
	if (i == 0
	    && wf->info
	    && !(wf->wf_flags & DCC_WF_PER_USER)
	    && dcc_clnt_wake_resolve())
		return DCC_WHITE_OK;

	if (!DCC_IS_TIME(now, wf->broken, DCC_WHITE_PARSE_DELAY)) {
		dcc_pemsg(EX_DATAERR, emsg, "%s still broken", wf->ascii_nm);
		if (i == 0)
			return (dcc_clnt_debug > 2
				? DCC_WHITE_CONTINUE : DCC_WHITE_OK);
		return (dcc_clnt_debug > 2
			? DCC_WHITE_COMPLAIN : DCC_WHITE_SILENT);
	}

	if (emsg && *emsg != '\0'
	    && (i < 0 || dcc_clnt_debug))
		dcc_error_msg("%s", emsg);

	i = make_white_hash(emsg, wf);
	if (i == 1)			/* all is good */
		return DCC_WHITE_OK;
	if (i == 0)			/* syntax error or bad hostname */
		return DCC_WHITE_CONTINUE;

	wf->broken = now + DCC_WHITE_PARSE_DELAY;
	return DCC_WHITE_COMPLAIN;	/* broken */
}



static u_char
lookup_white(DCC_EMSG emsg, DCC_TGTS *tgtsp, DCC_WF *wf,
	     const DCC_GOT_CKS *cks, const DCC_GOT_SUM *g)
{
	const DCC_WHITE_ENTRY *e;
	const DCC_WHITE_CIDR_ENTRY *cidrp;
	int bits;

	e = find_white(emsg, wf, g->type, g->sum, 0);
	if (e == FIND_WHITE_BROKEN) {
		*tgtsp = DCC_TGTS_OK;
		return 0;
	}

	if (!e) {
		if (g->type != DCC_CK_IP) {
			*tgtsp = DCC_TGTS_INVALID;
			return 1;
		}

		/* if we had no hit and it is an IP address,
		 * check the CIDR blocks */
		bits = 0;
		cidrp = &wf->info->hdr.cidr.e[wf->info->hdr.cidr.len];
		while (cidrp != wf->info->hdr.cidr.e) {
			--cidrp;
			/* look for the longest match */
			if (cidrp->bits <= bits)
				continue;
			if (DCC_IN_BLOCK(cks->ip_addr,
					 cidrp->addr, cidrp->mask)) {
				*tgtsp = cidrp->tgts;
				bits = cidrp->bits;
			}
		}
		if (bits == 0)
			*tgtsp = DCC_TGTS_INVALID;
		return 1;
	}

	*tgtsp = e->tgts;
	return 1;
}



/* check a local whitelist for a single checksum
 *	on exit the file is locked except after an error */
DCC_WHITE_RESULT
dcc_white_sum(DCC_EMSG emsg, DCC_WF *wf,
	      DCC_CK_TYPES type, const DCC_SUM sum,
	      DCC_WHITE_LISTING *listingp)
{
	DCC_WHITE_ENTRY *e;
	DCC_WHITE_RESULT result;

	result = dcc_rdy_white(emsg, wf);
	switch (result) {
	case DCC_WHITE_OK:
	case DCC_WHITE_CONTINUE:
		break;
	case DCC_WHITE_SILENT:
	case DCC_WHITE_NOFILE:
	case DCC_WHITE_COMPLAIN:
		*listingp = DCC_WHITE_RESULT_FAILURE;
		return result;
	}

	e = find_white(emsg, wf, type, sum, 0);
	if (e == FIND_WHITE_BROKEN) {
		*listingp = DCC_WHITE_RESULT_FAILURE;
		return DCC_WHITE_COMPLAIN;
	}

	if (!e)
		*listingp = DCC_WHITE_UNLISTED;
	else if (e->tgts == DCC_TGTS_OK2)
		*listingp = DCC_WHITE_HALF_LISTED;
	else if (e->tgts == DCC_TGTS_OK)
		*listingp = DCC_WHITE_LISTED;
	else if (e->tgts == DCC_TGTS_TOO_MANY)
		*listingp = DCC_WHITE_BLACK;
	else
		*listingp = DCC_WHITE_UNLISTED;

	return result;
}



/* see if an IP addess is that of one of our MX servers */
u_char
dcc_white_mx(DCC_EMSG emsg, DCC_TGTS *tgtsp, DCC_WF *wf, const DCC_GOT_CKS *cks)
{
	u_char result;

	result = 1;
	switch (dcc_rdy_white(emsg, wf)) {
	case DCC_WHITE_OK:
		break;
	case DCC_WHITE_CONTINUE:
		result = 0;
		break;
	case DCC_WHITE_NOFILE:
		*tgtsp = 0;
		return 1;
	case DCC_WHITE_SILENT:
	case DCC_WHITE_COMPLAIN:
		*tgtsp = 0;
		return 0;
	}

	if (cks->sums[DCC_CK_IP].type != DCC_CK_IP) {
		*tgtsp = 0;
		return result;
	}

	if (!lookup_white(emsg, tgtsp, wf, cks, &cks->sums[DCC_CK_IP]))
		return 0;

	if (*tgtsp != DCC_TGTS_OK_MX
	    && *tgtsp != DCC_TGTS_OK_MXDCC)
		*tgtsp = 0;
	return result;
}



/* See what a local whitelist file says about the checksums for a message.
 *	The message is whitelisted if at least one checksum is in the local
 *	whitelist or if there are two or more OK2 values.
 *	Otherwise it is blacklisted if at least one checksum is. */
DCC_WHITE_RESULT
dcc_white_cks(DCC_EMSG emsg, DCC_WF *wf,
	      const DCC_GOT_CKS *cks,	/* these checksums */
	      DCC_CKS_WTGTS wtgts,	/* whitelist targets */
	      DCC_WHITE_LISTING *listingp)
{
	const DCC_GOT_SUM *g;
	DCC_TGTS tgts, prev_tgts;
	DCC_WHITE_RESULT result;

	result = dcc_rdy_white(emsg, wf);
	switch (result) {
	case DCC_WHITE_OK:
	case DCC_WHITE_CONTINUE:
		break;
	case DCC_WHITE_NOFILE:
	case DCC_WHITE_COMPLAIN:
	case DCC_WHITE_SILENT:
		*listingp = DCC_WHITE_RESULT_FAILURE;
		return result;
	}

	/* look for each checksum in the hash file */
	*listingp = DCC_WHITE_UNLISTED;
	prev_tgts = tgts = DCC_TGTS_INVALID;
	for (g = &cks->sums[DCC_CK_TYPE_FIRST]; g <= LAST(cks->sums); ++g) {
		/* ignore checksums we don't have */
		if (g->type == DCC_CK_INVALID)
			continue;

		if (!lookup_white(emsg, &tgts, wf, cks, g)) {
			*listingp = DCC_WHITE_RESULT_FAILURE;
			return DCC_WHITE_COMPLAIN;
		}
		if (tgts == DCC_TGTS_INVALID)
			continue;

		if (wtgts)
			wtgts[g - cks->sums] = tgts;

		if (tgts == DCC_TGTS_OK) {
			/* found the checksum in our whitelist,
			 * so we have the answer */
			*listingp = DCC_WHITE_LISTED;

		} else if (tgts == DCC_TGTS_OK2) {
			if (prev_tgts == DCC_TGTS_OK2) {
				/* two half-white checksums count the same
				 * as a single pure white checksum
				 * and gives the answer */
				*listingp = DCC_WHITE_LISTED;
				continue;
			}
			prev_tgts = DCC_TGTS_OK2;

		} else if (tgts == DCC_TGTS_TOO_MANY) {
			if (*listingp == DCC_WHITE_UNLISTED)
				*listingp = DCC_WHITE_BLACK;
		}
	}

	return result;
}
